class ModguardError(Exception): ...


class ModguardParseError(ModguardError): ...


class ModguardSetupError(ModguardError): ...


__all__ = ["ModguardError", "ModguardParseError", "ModguardSetupError"]



# pyright: reportUnknownParameterType=false, reportMissingTypeArgument=false
# pyright: reportUnknownVariableType=false, reportUnknownArgumentType=false

from modguard.colors import BCOLORS
from modguard.core import ModuleTrie
from typing import Any, Dict, Tuple, Union

# This type hint only works on more recent versions
# result_dict: TypeAlias = dict[str, str | bool | 'result_dict']


def boundary_trie_to_dict(boundary_trie: ModuleTrie) -> Dict[str, Any]:
    result: Dict[str, Any] = dict()
    for node in boundary_trie:
        path = node.full_path
        if path == "":
            continue
        sections = path.split(".")
        current: Dict[str, Any] = result
        for section in sections:
            if section not in current:
                current[section] = dict()
            current = current[section]
        current["is_boundary"] = True

    return result


def dict_to_str(dict_repr: Dict[str, Any]) -> str:
    str_repr = ""

    def _recurs_build_string(str_repr: str, level: int, current: Dict[str, Any]) -> str:
        for k, v in current.items():
            if isinstance(v, dict):
                is_boundary = "is_boundary" in v.keys()
                str_repr += BCOLORS.ENDC + BCOLORS.ENDC + "\n" + "  " * level
                if is_boundary:
                    str_repr += BCOLORS.BOLD + "[B]"
                str_repr += k
                next_dict: Dict[str, Any] = v
                str_repr = _recurs_build_string(str_repr, level + 1, next_dict)
        return str_repr

    return _recurs_build_string(str_repr, 0, dict_repr) + "\n"


def dict_to_yaml(
    data: Union[dict, list, str, int, float, bool], indent: int = 0
) -> str:
    """
    Recursively converts a Python dictionary or list to a YAML-formatted string.

    Args:
        data (Union[dict, list, str, int, float, bool]): The data to convert to YAML.
        indent (int): The current indentation level (used for recursive calls).

    Returns:
        str: A string formatted as YAML.
    """
    yaml_str = ""
    if isinstance(data, dict):
        for key, value in data.items():
            yaml_str += " " * indent + str(key) + ":"
            if isinstance(value, (dict, list)):
                yaml_str += "\n" + dict_to_yaml(value, indent + 2)
            else:
                yaml_str += (
                    " " + str(value) + "\n"
                    if not isinstance(value, bool)
                    else " true\n"
                    if value
                    else " false\n"
                )
    elif isinstance(data, list):
        for item in data:
            yaml_str += " " * indent + "- "
            if isinstance(item, (dict, list)):
                yaml_str += "\n" + dict_to_yaml(item, indent + 2).lstrip()
            else:
                yaml_str += str(item) + "\n"
    else:
        yaml_str = " " * indent + str(data) + "\n"

    return yaml_str


def show(boundary_trie: ModuleTrie, write_file: bool = False) -> Tuple[str, str]:
    dict_repr = boundary_trie_to_dict(boundary_trie)
    yaml_result = dict_to_yaml(dict_repr)
    pretty_str_result = dict_to_str(dict_repr)
    if write_file:
        with open("modguard.yaml", "w") as f:
            f.write(yaml_result)
    return yaml_result, pretty_str_result



import os
from typing import Iterable, Optional

import yaml

from modguard import filesystem as fs
from modguard.check import check
from modguard.constants import CONFIG_FILE_NAME
from modguard.core import ScopeDependencyRules
from modguard.errors import ModguardError
from modguard.parsing import parse_project_config


def update_project_config(root: str, tags: set[str]):
    current_dir = os.getcwd()
    try:
        fs.chdir(root)
        project_config = parse_project_config()
        check_errors = check(
            root,
            project_config=project_config,
            exclude_paths=project_config.exclude,
        )
        for error in check_errors:
            if error.is_tag_error:
                invalid_tags = set(error.invalid_tags)
                existing_dependencies = set(
                    project_config.constraints.get(
                        error.source_tag, ScopeDependencyRules(depends_on=[])
                    ).depends_on
                )
                if error.source_tag in tags:
                    # This is updating the config for a new tag
                    project_config.constraints[error.source_tag] = ScopeDependencyRules(
                        depends_on=list(existing_dependencies | invalid_tags)
                    )
                if invalid_tags & tags:
                    # This is updating the config for an existing tag
                    project_config.constraints[error.source_tag] = ScopeDependencyRules(
                        depends_on=list(existing_dependencies | (invalid_tags & tags))
                    )

        modguard_yml_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
        modguard_yml_content = yaml.dump(project_config.model_dump())
        fs.write_file(modguard_yml_path, modguard_yml_content)

        check_errors = check(
            root, project_config=project_config, exclude_paths=project_config.exclude
        )
        if check_errors:
            return (
                "Could not auto-detect all dependencies, "
                "use 'modguard check' to finish initialization manually."
            )
    except Exception as e:
        fs.chdir(current_dir)
        raise e
    fs.chdir(current_dir)


def add_modules(paths: set[str], tags: Optional[set[str]]) -> Iterable[str]:
    new_tags: set[str] = set()
    # Validate paths
    for path in paths:
        fs.validate_path_for_add(path=path)
    # Build modules
    for path in paths:
        new_tag = fs.build_module(path=path, tags=tags)
        if new_tag:
            new_tags.add(new_tag)
    # Update project config
    project_root = fs.find_project_config_root(path=".")
    if not project_root:
        raise ModguardError(f"{CONFIG_FILE_NAME} not found.")
    warning = update_project_config(root=project_root, tags=tags if tags else new_tags)
    if warning:
        return [warning]
    return []



import os
from datetime import datetime
from typing import Optional


from modguard.constants import MODULE_FILE_NAME, CONFIG_FILE_NAME
from modguard.errors import ModguardError
from modguard.filesystem.project import find_project_config_root


def validate_module_config(root: str = ".") -> Optional[str]:
    file_path = os.path.join(root, f"{MODULE_FILE_NAME}.yml")
    if os.path.exists(file_path):
        return file_path
    file_path = os.path.join(root, f"{MODULE_FILE_NAME}.yaml")
    if os.path.exists(file_path):
        return file_path
    return


def validate_path_for_add(path: str) -> None:
    if not os.path.exists(path):
        raise ModguardError(f"{path} does not exist.")
    if os.path.isdir(path):
        if os.path.exists(
            os.path.join(path, f"{MODULE_FILE_NAME}.yml")
        ) or os.path.exists(os.path.join(path, f"{MODULE_FILE_NAME}.yaml")):
            raise ModguardError(f"{path} already contains a {MODULE_FILE_NAME}.yml")
        if not os.path.exists(os.path.join(path, "__init__.py")):
            raise ModguardError(
                f"{path} is not a valid Python package (no __init__.py found)."
            )
    # this is a file
    else:
        if not path.endswith(".py"):
            raise ModguardError(f"{path} is not a Python file.")
        if os.path.exists(path.removesuffix(".py")):
            raise ModguardError("{path} already has a directory of the same name.")
    root = find_project_config_root(path)
    if not root:
        raise ModguardError(
            f"{CONFIG_FILE_NAME} does not exist in any parent directories"
        )


def build_module(path: str, tags: Optional[set[str]]) -> Optional[str]:
    dirname = path.removesuffix(".py")
    new_tag = os.path.basename(dirname)
    if not tags:
        tags_to_write = [new_tag]
    else:
        tags_to_write = tags
    if os.path.isfile(path):
        # Create the package directory
        os.mkdir(dirname)
        # Write the __init__
        with open(f"{dirname}/__init__.py", "w") as new_init:
            new_init.write(f"""# Generated by modguard  on {datetime.now().strftime(
                '%Y-%m-%d %H:%M:%S')}
from .main import *
            """)
        # Move and rename the file
        os.rename(path, f"{dirname}/main.py")
    # Write the module.yml
    with open(f"{dirname}/{MODULE_FILE_NAME}.yml", "w") as f:
        f.write(f"tags: [\"{'","'.join(tags_to_write)}\"]\n")
    if not tags:
        return new_tag



from modguard.core.config import ModuleConfig, ScopeDependencyRules, ProjectConfig
from modguard.core.module import ModuleNode, ModuleTrie

__all__ = [
    "ModuleConfig",
    "ScopeDependencyRules",
    "ProjectConfig",
    "ModuleNode",
    "ModuleTrie",
]



from typing import Optional

from modguard import filesystem as fs
from modguard.core import ModuleTrie
from modguard.parsing import parse_module_config, parse_interface_members


def build_module_trie(
    root: str,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
) -> ModuleTrie:
    boundary_trie = ModuleTrie()

    for dir_path in fs.walk_modules(
        root,
        exclude_paths=exclude_paths,
        exclude_hidden_paths=exclude_hidden_paths,
    ):
        module_config = parse_module_config(dir_path)
        if module_config is None:
            raise ValueError(f"Could not parse module config for {dir_path}")
        boundary_trie.insert(
            config=module_config,
            path=fs.file_to_module_path(dir_path),
            interface_members=parse_interface_members(dir_path),
        )

    return boundary_trie



from modguard.parsing.config import parse_module_config, parse_project_config
from modguard.parsing.imports import get_project_imports
from modguard.parsing.interface import parse_interface_members
from modguard.parsing.modules import build_module_trie


__all__ = [
    "parse_module_config",
    "parse_project_config",
    "get_project_imports",
    "parse_interface_members",
    "build_module_trie",
]



import os
from dataclasses import field, dataclass
from typing import Optional

import yaml

from modguard import errors
from modguard import filesystem as fs
from modguard.check import check
from modguard.constants import MODULE_FILE_NAME, CONFIG_FILE_NAME
from modguard.core import ProjectConfig, ScopeDependencyRules

__module_yml_template = """tags: ['{dir_name}']\n"""


@dataclass
class ModuleInitResult:
    module_paths: list[str] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)


def init_modules(
    root: str, depth: int, exclude_paths: Optional[list[str]] = None
) -> ModuleInitResult:
    module_paths: list[str] = []
    warnings: list[str] = []
    for dir_path in fs.walk_pypackages(root, depth=depth, exclude_paths=exclude_paths):
        module_yml_path = os.path.join(dir_path, f"{MODULE_FILE_NAME}.yml")
        module_paths.append(dir_path)
        if os.path.exists(module_yml_path):
            warnings.append(f"Module file '{module_yml_path}' already exists.")
            continue
        module_yml_content = __module_yml_template.format(
            dir_name=dir_path.replace(os.path.sep, ".")
        )
        fs.write_file(module_yml_path, module_yml_content)

    return ModuleInitResult(module_paths=module_paths, warnings=warnings)


@dataclass
class InitRootResult:
    warnings: list[str] = field(default_factory=list)


def init_root(root: str, exclude_paths: Optional[list[str]] = None) -> InitRootResult:
    project_config_path = fs.get_project_config_path(root)
    if project_config_path:
        return InitRootResult(
            warnings=[f"Project already contains {CONFIG_FILE_NAME}.yml"]
        )

    project_config = ProjectConfig()
    check_errors = check(
        root, project_config=project_config, exclude_paths=exclude_paths
    )
    for error in check_errors:
        if error.is_tag_error:
            existing_dependencies = set(
                project_config.constraints.get(
                    error.source_tag, ScopeDependencyRules(depends_on=[])
                ).depends_on
            )
            project_config.constraints[error.source_tag] = ScopeDependencyRules(
                depends_on=list(existing_dependencies | set(error.invalid_tags))
            )

    modguard_yml_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
    modguard_yml_content = yaml.dump(project_config.model_dump())
    fs.write_file(modguard_yml_path, modguard_yml_content)

    check_errors = check(
        root, project_config=project_config, exclude_paths=exclude_paths
    )
    if check_errors:
        return InitRootResult(
            warnings=[
                "Could not auto-detect all dependencies, use 'modguard check' to finish initialization manually."
            ]
        )

    return InitRootResult(warnings=[])


def init_project(
    root: str, depth: Optional[int] = None, exclude_paths: Optional[list[str]] = None
) -> list[str]:
    if not os.path.isdir(root):
        raise errors.ModguardSetupError(f"The path {root} is not a directory.")

    warnings: list[str] = []

    if depth is None:
        module_init_result = init_modules(root, depth=1, exclude_paths=exclude_paths)
        warnings.extend(module_init_result.warnings)
        if len(module_init_result.module_paths) == 1:
            result = init_modules(
                module_init_result.module_paths[0], depth=1, exclude_paths=exclude_paths
            )
            warnings.extend(result.warnings)
    else:
        module_init_result = init_modules(
            root, depth=depth, exclude_paths=exclude_paths
        )
        warnings.extend(module_init_result.warnings)

    init_root_result = init_root(root, exclude_paths=exclude_paths)
    warnings.extend(init_root_result.warnings)

    return warnings



import argparse
import sys
from typing import Optional

from modguard.add import add_modules
from modguard.check import check, ErrorInfo
from modguard import filesystem as fs
from modguard.init import init_project
from modguard.loading import stop_spinner, start_spinner
from modguard.parsing import parse_project_config, build_module_trie
from modguard.show import show
from modguard.colors import BCOLORS


def print_errors(error_list: list[ErrorInfo]) -> None:
    sorted_results = sorted(error_list, key=lambda e: e.location)
    for error in sorted_results:
        print(
            f"❌ {BCOLORS.FAIL}{error.location}{BCOLORS.WARNING}: {error.message}",
            file=sys.stderr,
        )


# todo default to full path from project root
# todo mutate module trie (insert)
#


def add_base_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-e",
        "--exclude",
        required=False,
        type=str,
        metavar="file_or_path,...",
        help="Comma separated path list to exclude. tests/, ci/, etc.",
    )


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="modguard",
        add_help=True,
        epilog="Make sure modguard is run from the root of your Python project,"
        " and `modguard.yml` is present",
    )
    subparsers = parser.add_subparsers(title="commands", dest="command")
    init_parser = subparsers.add_parser(
        "init",
        prog="modguard init",
        help="Initialize boundaries between top-level modules and write dependencies to "
        "`modguard.yml`",
        description="Initialize boundaries between top-level modules and write dependencies to "
        "`modguard.yml`",
    )
    add_base_arguments(init_parser)
    check_parser = subparsers.add_parser(
        "check",
        prog="modguard check",
        help="Check existing boundaries against your dependencies and module interfaces",
        description="Check existing boundaries against your dependencies and module interfaces",
    )
    add_base_arguments(check_parser)
    show_parser = subparsers.add_parser(
        "show",
        prog="modguard show",
        help="Show your existing boundaries",
        description="Show your existing boundaries",
    )
    add_base_arguments(show_parser)
    show_parser.add_argument(
        "-w",
        "--write",
        required=False,
        dest="write",
        action="store_true",
        default=False,
        help="Write the output to an `interface.yaml` file",
    )
    add_parser = subparsers.add_parser(
        "add",
        prog="modguard add",
        help="Create a new module boundary around an existing file or folder",
        description="Initialize boundaries between top-level modules and write dependencies to "
        "`modguard.yml`",
    )
    add_parser.add_argument(
        "path",
        type=str,
        metavar="file_or_path,...",
        help="The path(s) of the file or directory to create a module boundary around. "
        "Use a comma-separated list for multiple.",
    )
    add_parser.add_argument(
        "-t",
        "--tags",
        required=False,
        type=str,
        metavar="tag,...",
        help="The tag for the module to be initialized with."
        "Use a comma-separated list for multiple.",
    )
    return parser


def parse_arguments(args: list[str]) -> argparse.Namespace:
    parser = build_parser()
    parsed_args = parser.parse_args(args)

    if args[0] not in ["init", "add"]:
        fs.validate_project_config_path()

    return parsed_args


def modguard_check(
    exclude_paths: Optional[list[str]] = None,
):
    try:
        project_config = parse_project_config()
        if exclude_paths is not None and project_config.exclude is not None:
            exclude_paths.extend(project_config.exclude)
        else:
            exclude_paths = project_config.exclude
        result: list[ErrorInfo] = check(
            ".",
            project_config,
            exclude_paths=exclude_paths,
            exclude_hidden_paths=project_config.exclude_hidden_paths,
        )
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)

    stop_spinner()
    if result:
        print_errors(result)
        sys.exit(1)
    print(f"✅ {BCOLORS.OKGREEN}All modules safely guarded!")
    sys.exit(0)


def modguard_show(
    write_file: bool,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
):
    try:
        mt = build_module_trie(
            ".", exclude_paths=exclude_paths, exclude_hidden_paths=exclude_hidden_paths
        )
        _, pretty_result = show(mt, write_file=write_file)
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)
    stop_spinner()
    print(pretty_result)
    sys.exit(0)


def modguard_init(exclude_paths: Optional[list[str]] = None):
    try:
        warnings = init_project(root=".", exclude_paths=exclude_paths)
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)

    stop_spinner()
    if warnings:
        print("\n".join(warnings))
    print(f"✅ {BCOLORS.OKGREEN}Modguard initialized.")
    sys.exit(0)


def modguard_add(paths: set[str], tags: Optional[set[str]] = None) -> None:
    try:
        warnings = add_modules(paths, tags)
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)

    stop_spinner()
    if warnings:
        print("\n".join(warnings))
    print(f"✅ {BCOLORS.OKGREEN}Modguard initialized.")
    sys.exit(0)


def main() -> None:
    args = parse_arguments(sys.argv[1:])
    if args.command == "add":
        paths = set(args.path.split(","))
        tags = set(args.tags.split(",")) if args.tags else None
        modguard_add(paths=paths, tags=tags)
        return
    exclude_paths = args.exclude.split(",") if args.exclude else None
    if args.command == "init":
        start_spinner("Initializing...")
        modguard_init(exclude_paths=exclude_paths)
    elif args.command == "check":
        start_spinner("Scanning...")
        modguard_check(exclude_paths=exclude_paths)
    elif args.command == "show":
        start_spinner("Scanning...")
        modguard_show(write_file=args.write, exclude_paths=exclude_paths)
    else:
        print("Unrecognized command")
        exit(1)


if __name__ == "__main__":
    main()
